import ctypes

c_uint8 = ctypes.c_uint8
c_uint32 = ctypes.c_uint32

TestConfig = {
    "DevName": "Valerie",
    "Version": "v2.1.10",
    "ProductionDate": "20191106",
    "HolderName": "Tanja",
    "StuName": "Valerie",
}
