from bidict import bidict

MyToolItNetworkNr = bidict(
    {
        "BroadCast": 0,
        "STH1": 1,
        "STH2": 2,
        "STH3": 3,
        "STH4": 4,
        "STH5": 5,
        "STH6": 6,
        "STH7": 7,
        "STH8": 8,
        "STH9": 9,
        "STH10": 10,
        "STH11": 11,
        "STH12": 12,
        "STH13": 13,
        "STH14": 14,
        "SPU1": 15,
        "SPU2": 16,
        "STU1": 17,
        "STU2": 18,
        "STU3": 19,
        "STU4": 20,
        "STU5": 21,
        "STU6": 22,
        "STU7": 23,
        "STU8": 24,
        "STU9": 25,
        "STU10": 26,
        "STU11": 27,
        "STU12": 28,
        "STU13": 29,
        "STU14": 30,
        "BroadCastNoAck": 31,
    }
)

MyToolItNetworkName = {
    MyToolItNetworkNr["BroadCast"]: "Broadcast with ACK",
    MyToolItNetworkNr["STH1"]: "STH1",
    MyToolItNetworkNr["STH2"]: "STH2",
    MyToolItNetworkNr["STH3"]: "STH3",
    MyToolItNetworkNr["STH4"]: "STH4",
    MyToolItNetworkNr["STH5"]: "STH5",
    MyToolItNetworkNr["STH6"]: "STH6",
    MyToolItNetworkNr["STH7"]: "STH7",
    MyToolItNetworkNr["STH8"]: "STH8",
    MyToolItNetworkNr["STH9"]: "STH9",
    MyToolItNetworkNr["STH10"]: "STH10",
    MyToolItNetworkNr["STH11"]: "STH11",
    MyToolItNetworkNr["STH12"]: "STH12",
    MyToolItNetworkNr["STH13"]: "STH13",
    MyToolItNetworkNr["STH14"]: "STH14",
    MyToolItNetworkNr["SPU1"]: "SPU1",
    MyToolItNetworkNr["SPU2"]: "SPU2",
    MyToolItNetworkNr["STU1"]: "STU1",
    MyToolItNetworkNr["STU2"]: "STU2",
    MyToolItNetworkNr["STU3"]: "STU3",
    MyToolItNetworkNr["STU4"]: "STU4",
    MyToolItNetworkNr["STU5"]: "STU5",
    MyToolItNetworkNr["STU6"]: "STU6",
    MyToolItNetworkNr["STU7"]: "STU7",
    MyToolItNetworkNr["STU8"]: "STU8",
    MyToolItNetworkNr["STU9"]: "STU9",
    MyToolItNetworkNr["STU10"]: "STU10",
    MyToolItNetworkNr["STU11"]: "STU11",
    MyToolItNetworkNr["STU12"]: "STU12",
    MyToolItNetworkNr["STU13"]: "STU13",
    MyToolItNetworkNr["STU14"]: "STU14",
    MyToolItNetworkNr["BroadCastNoAck"]: "Broadcast WITHOUT ACK",
}
